/* 
 * Copyright (c) 2006, Karl Helgason
 * 
 * 2007/1/8 modified by p.j.leonard
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials
 *       provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package be.hogent.tarsos.dsp.constantq;


import java.io.File;
import java.io.IOException;

import javax.sound.sampled.UnsupportedAudioFileException;

import be.hogent.tarsos.dsp.AudioDispatcher;
import be.hogent.tarsos.dsp.AudioEvent;
import be.hogent.tarsos.dsp.AudioProcessor;
import be.hogent.tarsos.dsp.util.fft.FFT;

// Implementation of Constant Q Transform
//
// References:
//
// Judith C. Brown,
// Calculation of a constant Q spectral transform, J. Acoust. Soc. Am., 89(1):
// 425-434, 1991.
// see http://www.wellesley.edu/Physics/brown/pubs/cq1stPaper.pdf
//
// Judith C. Brown and MillerS. Puckette,
// An efficient algorithm for the calculation of a constant Q transform, J.
// Acoust. Soc. Am., Vol. 92, No. 5, November 1992
// see http://www.wellesley.edu/Physics/brown/pubs/effalgV92P2698-P2701.pdf
//
// Benjamin Blankertz,
// The Constant Q Transform
// see
// http://wwwmath1.uni-muenster.de/logik/org/staff/blankertz/constQ/constQ.pdf
//
// mods by p.j.leonard
// 0 - default centered quantized freq.
// 1 -
// 3 - j centered samples
// 4 - original

public class FFTConstantQ {

	double q; // Constant Q

	int k; // Number of output bands

	int fftlen; // FFT size

	float[] freqs;

	float[][] qKernel;

	int[][] qKernel_indexes;

	FFT fft;


	public FFT getFFT() {
		return fft;
	}

	public float[] getFreqs() {
		return freqs;
	}
	public int getFFTSize() {
		return fftlen;
	}

	public int getNumberOfOutputBands() {
		return k;
	}

	float sampleRate = 44100;

	float minFreq = 100;

	float maxFreq = 3000;

	float binsPerOctave = 48;

	float threshold = 0.001f; // Lower number, better quality !!! 0 = best

	float spread = 1.0f;

	String kernelsident;


	public FFTConstantQ(float sampleRate, float minFreq, float maxFreq,
			float binsPerOctave) {
		this(sampleRate,minFreq,maxFreq,binsPerOctave,0.001f,1.0f);
	}

	public FFTConstantQ(float sampleRate, float minFreq, float maxFreq,
			float binsPerOctave, float threshold,float spread) {
		this.sampleRate = sampleRate;
		this.minFreq = minFreq;
		this.maxFreq = maxFreq;
		this.binsPerOctave = binsPerOctave;
		this.threshold = threshold;
		this.spread = spread;
		init();
	}

	

	private void init() {

		// Calculate Constant Q
	//	System.out.println("Spread = " + spread + " Key:" + key);
		
		q = 1.0 / (Math.pow(2, 1.0 / binsPerOctave) - 1.0) / spread;

		// Calculate number of output bins
		k = (int) Math.ceil(binsPerOctave * Math.log(maxFreq / minFreq)
				/ Math.log(2));

		// Calculate length of FFT
		float calc_fftlen = (float) Math.ceil(q * sampleRate / minFreq);
		fftlen = (int) Math.pow(2, Math.ceil(Math.log(calc_fftlen)
				/ Math.log(2)));

		// Create FFT object
		fft = new FFT(fftlen);
		qKernel = new float[k][];
		qKernel_indexes = new int[k][];
		freqs = new float[k];

		// Calculate Constant Q kernels

		float[] temp = new float[fftlen * 2];
		float[] ctemp = new float[fftlen * 2];
		int[] cindexes = new int[fftlen];

		for (int i = 0; i < k; i++) {
			float[] sKernel = temp;
			// Calculate the frequency of current bin
			freqs[i] = (float) (minFreq * Math.pow(2, i / binsPerOctave));

			float len = (float) (q * sampleRate / freqs[i]);

			// float halflen= len*0.5;
			// window is symmetric around center point of frame
			// calculate second half of the kernel

			for (int j = 0; j < fftlen / 2; j++) {

				float aa;

				aa = (float) (j + 0.5) / len;

				if (aa < .5) {
					float a = (float) (2.0 * Math.PI * aa);
					float window = (float) (0.5 * (1.0 + Math.cos(a))); // Hanning

					window /= len;

					// Calculate kernel
					float x = (float) (2.0 * Math.PI * freqs[i] * (j + 0.5D) / sampleRate);

					sKernel[fftlen + j * 2] = (float) (window * Math.cos(x));
					sKernel[fftlen + j * 2 + 1] = (float) (window * Math.sin(x));
				} else {
					sKernel[fftlen + j * 2] = 0.0f;
					sKernel[fftlen + j * 2 + 1] = 0.0f;
				}
			}

			// reflect to genereate first half

			int halfway = fftlen / 2;

			for (int j = 0; j < halfway; j++) {
				int i1 = halfway - j - 1;
				int i2 = halfway + j;

				sKernel[i1 * 2] = sKernel[2 * i2];
				sKernel[i1 * 2 + 1] = -sKernel[2 * i2 + 1];
			}

			// Perform FFT on kernel
			fft.complexForwardTransform(sKernel);

			// Remove all zeros from kernel to improve performance
			float[] cKernel = ctemp;

			int k = 0;
			for (int j = 0, j2 = sKernel.length/2 -1; j < sKernel.length / 2; j += 2, j2 -= 2) {
				float absval = fft.modulus(cKernel, j);
				absval += fft.modulus(cKernel, j2);
				if (absval > threshold) {
					cindexes[k] = j;
					cKernel[2 * k] = sKernel[j] + sKernel[j2];
					cKernel[2 * k + 1] = sKernel[j + 1] + sKernel[j2 + 1];
					k++;
				}
			}

			sKernel = new float[k * 2];
			int[] indexes = new int[k];

			for (int j = 0; j < k * 2; j++)
				sKernel[j] = cKernel[j];
			for (int j = 0; j < k; j++)
				indexes[j] = cindexes[j];

			// Normalize fft output
			for (int j = 0; j < sKernel.length; j++)
				sKernel[j] /= fftlen;

			// Perform complex conjugate on sKernel
			for (int j = 1; j < sKernel.length; j += 2)
				sKernel[j] = -sKernel[j];

			qKernel_indexes[i] = indexes;
			qKernel[i] = sKernel;

		}
	}

	/**
     *
     *  take a buff_in of samples and calculate the constant Q coeffs.
     *
     *
     * @param buff_in
     * @param buff_out
     */
	public void calc(float[] buff_in, float[] buff_out) {
		fft.forwardTransform(buff_in);
		for (int i = 0; i < qKernel.length; i++) {
			float[] kernel = qKernel[i];
			int[] indexes = qKernel_indexes[i];
			float t_r = 0;
			float t_i = 0;
			for (int j = 0, l = 0; j < kernel.length; j += 2, l++) {
				int jj = indexes[l];
				float b_r = buff_in[jj];
				float b_i = buff_in[jj + 1];
				float k_r = kernel[j];
				float k_i = kernel[j + 1];
				// COMPLEX: T += B * K
				t_r += b_r * k_r - b_i * k_i;
				t_i += b_r * k_i + b_i * k_r;
			}
			buff_out[i * 2] = t_r;
			buff_out[i * 2 + 1] = t_i;
		}
	}

	public int getFFTlength() {
		return fftlen;
	}
	
	public static void main(String...strings) throws UnsupportedAudioFileException, IOException{
		AudioDispatcher d = AudioDispatcher.fromFile(new File("/home/joren/Dropbox/0110.be/Panako/Panako-Java/etc/shifted/01. County Line_real_noise_52s-62s_youtube.wav"), 16384, 16384-256);
		d.addAudioProcessor(new AudioProcessor(){
			FFTConstantQ fftConstantQ = new FFTConstantQ(8000, 64, 1024, 48);
			@Override
			public void processingFinished() {
			}
			
			@Override
			public boolean process(AudioEvent audioEvent) {
				float[] buffer = audioEvent.getFloatBuffer();
				float[] out = new float [400];
				fftConstantQ.calc(buffer, out);
				for(int i = 0 ; i<out.length;i++){
					System.out.println(out[i]);
				}
				return false;
			}
		});
		d.run();
	}
}

/*
-3.894599299307797E-5
-0.002364872673562526
-0.001269957294900326
0.00241431429712889
8.162188812993659E-4
0.0015097339426078944
0.004086458892626148
-0.0011176003201098874
-3.500687749983282E-4
-0.0014710443525738975
-0.0022831456890600554
-6.897295134546075E-4
0.0012446329076156674
7.578098174103016E-4
-5.963412747476313E-4
-3.3232866083997194E-4
-3.219517219640785E-4
-6.972252851806101E-4
0.0010479214622563385
1.034778086203617E-4
-3.008540351817123E-4
-2.404227368654442E-4
0.0015235895711848786
3.847545090018413E-4
-3.1704998355912024E-4
3.347959427858165E-5
-0.001085811305966813
-9.262882214758434E-4
-4.64518489264564E-4
4.6598747584489485E-4
-0.001170088753703323
-5.815385984591142E-4
-2.599780626203087E-4
-0.0013021945339479787
-0.00102341628630602
-4.253175005343908E-4
-6.440816891598852E-4
-7.128304151917579E-4
4.978571235163532E-4
-3.415923298220958E-4
0.002034303923076684
-8.377581832708859E-4
0.002427159042550249
-7.751171670892966E-4
7.740896713634129E-4
-7.056797064312978E-4
1.7862236646497455E-4
-0.001365561616337489
5.6689837240381E-4
-0.0010273250752575085
-8.953888997693065E-4
-6.623695372758512E-4
-9.658686892428267E-4
-0.0016914335343165203
2.8917739896825763E-4
0.002411056107407969
-0.0021883990518195137
0.004448118574261501
-0.0013002848780284543
5.014570073072222E-4
2.6682674406661974E-4
-0.0010675004213595652
-4.419086144645328E-4
-0.0024712718054582155
-6.99398786152073E-4
-0.0022548331203567588
-0.0012683145861029795
-5.253633138330127E-4
-2.7409461472387907E-4
-4.979680816800236E-4
0.0029193238908643737
-0.0017460219427657148
-0.0029372963435109344
-0.008081259605836293
-0.012451783967161841
-0.034929201730889314
-0.0020450064699523954
-0.046289335609828126
0.00321710436196212
-0.02018065455586473
-0.003953168090368777
-0.0015720824139709632
-0.00163047715882302
0.0012054709833231832
0.002258244097077973
0.0012550569621997579
-4.6689729392467515E-4
-2.7143465005892377E-4
-0.002465093287767051
-3.481015062381998E-4
-4.148356293667277E-4
-2.2835061867252846E-4
4.963102380046237E-4
-7.25595854714383E-4
-6.70229827385771E-4
4.0028496830003346E-4
-6.230417502538551E-4
1.629040709628894E-4
-9.06663885854381E-4
-1.6734870174289778E-4
-0.0011736925253900406
9.556914923129328E-4
-2.7175294148820613E-4
-1.3281085200140324E-4
5.509615418526953E-4
-0.001173146136707371
4.2908768982481856E-4
-9.218054893567985E-4
-3.7765358037134667E-4
-8.690737883254704E-4
6.556433836153009E-4
-0.0011032385833400324
0.001842034609447838
-0.0015983920090172344
8.005338073789833E-4
-0.001711412413518058
-5.516673261153693E-4
-6.490000749038062E-4
-3.9502691970321596E-4
0.0011112038779096123
0.0014753162358097222
0.0018499310885429672
0.0019799782881270048
5.758195446701159E-4
-5.983222315681443E-4
-9.565010258705031E-4
-0.0028068248270237844
0.0013002722752184432
0.001594137314612069
0.003488148051812028
0.008953225245171954
-0.002365267157407929
0.007894780689676358
-7.932269040731828E-4
0.0035040571922114916
0.004974142692431754
0.00187711891852137
7.938562393045139E-4
-2.3989776801090463E-6
-0.0020305267108759553
-0.001054999091020753
-3.0025239014407816E-4
-6.047419116559307E-4
-4.8881142730787227E-5
-2.0845152674474208E-4
-2.5316414966153046E-4
1.5525603557356392E-4
-2.5518178095542207E-4
-3.58116071362643E-4
-4.236441714775562E-4
-0.0010984571598722642
6.313237973390006E-4
-2.4398115577406974E-4
0.0015788733812646762
9.279554154394122E-4
8.271471096056077E-4
0.0010428605647752212
6.7531556973796E-5
3.8500443322763003E-4
2.6346481408903553E-4
3.7797910821309015E-4
5.671621946658937E-4
8.43466104231331E-4
4.782483923051348E-5
3.9380848173040594E-4
-6.749338130114135E-4
-2.0940834043322145E-4
-3.6405850587326505E-4
5.37324709733875E-4
-8.611603900034961E-4
0.0029496025773599116
0.002313795723931795
0.0017125939448097418
0.007058634112054971
-0.0010352463312579823
0.003906706161485341
-8.252430127554781E-4
-4.67356929112579E-4
-7.135614235466526E-4
-8.845866404849755E-4
-0.001867840996197548
-4.798599568019663E-4
-0.0026087674877907302
-0.001173547538597052
0.0015826194181428842
-7.614271705513382E-5
0.008221422659513202
0.002371527891743279
0.006742363593855002
8.198882931707964E-4
0.0018381110205409955
-8.344382550535477E-4
3.9300144200242655E-4
-1.9791206895658424E-4
2.7680724352775315E-4
5.195715318475823E-4
3.517144521851751E-4
2.831195883278921E-4
-1.8365520599902764E-4
-4.377230977073527E-4
-0.0013208119359280344
-0.0019818272746163993
-0.005672488771661177
-0.00853673585052788
-0.00722588842189553
-0.010865719086648889
-0.0017540165671793661
-0.004334201157083619
0.0013262260706741243
-6.203748108742401E-4
7.947505397629368E-4
-6.050019994550349E-4
6.242520936520265E-4
2.968797319640439E-5
6.336443982459859E-4
1.7441543905932525E-4
3.256875458277224E-4
-4.0691604165814076E-4
-5.310223782673029E-5
-3.587797176150424E-4
3.717473211765073E-5
-1.4961128794181084E-4
-1.765538918614902E-4
-9.98786344174054E-5
-3.8082695716458117E-4
3.4785855147481784E-5
-3.1195856010595113E-4
3.268008489430215E-4
-3.6309059578775246E-4
0.0010612304555677116
-4.2836689636907907E-4
6.560023398918992E-4
-8.789709741819905E-5
1.0387795840282083E-5
1.5555991289735422E-4
5.029969247531463E-5
-1.1438495689658937E-4
5.35752603896849E-5
-1.878689054657461E-4
1.102374299445342E-4
-2.388641602480218E-5
8.861122148185005E-5
1.277998578591841E-4
7.63925242417448E-6
1.5063444200432859E-4
-2.3709110494466373E-5
5.536342001231927E-5
2.3010008902624863E-6
8.003590212289127E-5
1.0875937033203107E-6
1.1197795890515908E-4
7.242793791186596E-5
9.744559264743402E-5
1.387138037332266E-4
6.909046277054757E-5
1.1866233578698961E-4
2.5665716660864475E-5
1.0946667213711864E-4
2.3240992726837754E-5
1.4374666382669458E-4
7.132280447616615E-5
1.89396561004508E-4
1.8229640145793385E-4
1.1331263436633793E-4
2.3747615425919976E-4
9.586641320436248E-5
3.2716572301719015E-4
2.753514828724832E-4
3.365184948960462E-4
4.0300112289333684E-4
8.714326940848202E-5
4.7976763082880876E-4
-2.144399526217325E-5
3.9096274385134837E-4
8.222108525980162E-5
1.2274813303387992E-4
1.8789551705685403E-4
4.932460480278358E-4
3.6749888873255414E-4
-4.541756576903045E-4
0.0010017426784345348
0.00141732566427547
0.0015965009592601577
0.007763747268784844
0.001375741505617549
0.007214681105842505
6.346830781403429E-4
0.002402894023282273
-1.5936231075046531E-4
7.446544460187894E-4
-2.1791487196102637E-4
1.0326962070642808E-5
2.518753799864824E-4
-1.6800150715300322E-4
6.481491478848705E-5
1.8002711672975122E-4
-3.442871559198652E-4
5.721543369734489E-4
0.0023558385575171478
5.926426846326932E-4
0.0044019730267949005
-3.9735398833879674E-4
0.0019121472930096971
-4.6783399329004664E-4
-3.7630012666480054E-4
1.933584475271695E-4
-0.001129706467694916
-8.61179246320491E-4
-8.697091876234984E-4
-0.0014025635181761783
-6.60808591415423E-5
-6.976128433416719E-4
-1.4088214281363052E-4
-5.560038814694535E-4
-3.6120745128217103E-4
-8.593471092323076E-5
-2.3029726521546227E-4
3.6892655216411743E-4
-5.919008507286778E-5
4.640483791958055E-5
-1.5570685502426182E-4
-1.7344187804897352E-4
-3.662523019292263E-4
-2.4603610660206906E-4
-2.3680706669974017E-4
-2.21349296037291E-4
-1.1022941659166926E-4
-2.6854177599420424E-4
-2.1652454767010084E-4
-1.5701387884324298E-4
-6.045361573254865E-4
-1.6562894882593427E-4
-3.929190257468106E-4
-9.685286773405551E-4
-0.0018263051775829287
-0.001146885966504297
1.6857762139446526E-4
-0.0011837597273439307
0.013287047760343675
-0.001880684139712317
0.014826707681863205
-0.0031734926416266945
0.00335424699218128
-0.0026597743312529397
-3.701592878775851E-4
-1.9916796154584277E-4
1.899342375031185E-4
6.643338872439152E-4
-2.5070922074384334E-4
3.02554457968736E-4
-2.4129689686273417E-4
7.38266080478907E-5
2.2121560903257093E-4
-6.21906608178369E-5
5.160325686728054E-4
-2.9847330076646127E-4
1.6814032487591336E-4
-4.2613663198703337E-4
-2.2178155207613033E-4
7.723088919736961E-7
-4.6967027244606755E-4
6.037891204594095E-4
-7.217931112168883E-4
0.001058799958232708
-4.3101819105003866E-4
0.0010751397561138366
2.430536066150256E-5
6.658722162387007E-4
2.670493993177661E-4
4.7616043051183706E-4
5.58114333213994E-4
5.477463310110304E-4
4.0129381939331654E-4
3.549273798173298E-4
1.9080205182149074E-4
9.494708789018853E-4
-0.001520673074707107
2.6076842387321364E-4
-0.004660992761191063
-0.007752453910578448
-0.005164692612676043
-0.009693720354084995
-0.002137440616361886
-0.0022974932683262517

 * */